from typing import Generator
import litellm

import logging
import litellm

from aiconsole.code_interpreters.language_map import language_map
from aiconsole.gpt.parse_partial_json import parse_partial_json
from aiconsole.gpt.merge_deltas import merge_deltas

log = logging.getLogger(__name__)

function_schema = {
  "name": "execute",
  "description":
  "Executes code on the user's machine, **in the users local environment**, and returns the output",
  "parameters": {
    "type": "object",
    "properties": {
      "language": {
        "type": "string",
        "description":
        "The programming language (required parameter to the `execute` function)",
        "enum": [language for language in language_map.keys()]
      },
      "code": {
        "type": "string",
        "description": "The code to execute (required)"
      }
    },
    "required": ["language", "code"]
  },
}


def coding_llm(model, max_tokens, temperature, api_key, api_base, messages) -> Generator[dict, None, None]:
    
    log.debug("Sending this to the OpenAI LLM: %s", messages)

    params = {
        'model': model,
        'messages': messages,
        'stream': True,
        'functions': [function_schema]
    }

    if api_base:
        params["api_base"] = api_base
    if api_key:
        params["api_key"] = api_key
    if max_tokens:
        params["max_tokens"] = max_tokens
    if temperature: # TODO: Bug, what if it's 0?
        params["temperature"] = temperature
    
    log.debug("Sending this to LiteLLM: %s", params)

    response = litellm.completion(**params)

    accumulated_deltas = {}
    language = None
    code = ""

    for chunk in response:
        if ('choices' not in chunk or len(chunk['choices']) == 0):
            continue

        delta = chunk["choices"][0]["delta"]

        # Accumulate deltas
        accumulated_deltas = merge_deltas(accumulated_deltas, delta)
        log.debug("Accumulated deltas: %s", accumulated_deltas)

        if "content" in delta and delta["content"]:
            yield {"message": delta["content"]}

        
        if "function_call" in accumulated_deltas and "arguments" in accumulated_deltas["function_call"]:
            # This can now be both a string and a json object
            arguments = parse_partial_json(accumulated_deltas["function_call"]["arguments"])
            languages = language_map.keys()
            
            # We need to handle incorrect OpenAI responses, sometmes arguments is a string containing the code
            if isinstance(arguments, str):
                if language is None and "name" in accumulated_deltas["function_call"] and accumulated_deltas["function_call"]["name"] in languages:
                    language = accumulated_deltas["function_call"]["name"]
                    yield {"language": language}

                if arguments and not arguments.startswith("{"):
                    if language is None:
                        language = "python"
                        yield {"language": language}

                    code_delta = arguments[len(code):]
                    code = arguments
                    
                    if code_delta:
                        yield {"code": code_delta}

            else:
                if language is None and arguments and "language" in arguments and arguments["language"] in languages:
                    language = arguments["language"]
                    yield {"language": language}

                if language is not None and arguments and "code" in arguments:
                    code_delta = arguments["code"][len(code):]
                    code = arguments["code"]
                    
                    if code_delta:
                        yield {"code": code_delta}
                        
