import importlib.util
import logging
import os
import re
from typing import Dict

import watchdog.events
import watchdog.observers

from aiconsole.aic_types import Material
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.settings import MATERIALS_CORE_RESOURCE, MATERIALS_DIRECTORY
from aiconsole.materials.documentation_from_code import documentation_from_code
from aiconsole.utils.connection_manager import NotificationWSMessage, connection_manager
from aiconsole.utils.list_files_in_file_system import list_files_in_file_system
from aiconsole.utils.list_files_in_resource_path import list_files_in_resource_path

log = logging.getLogger(__name__)


class Materials:
    """
    Materials' class is for managing the .md and .py material files.
    """

    materials: Dict[str, Material]

    def __init__(self):
        self.materials = {}

        observer = watchdog.observers.Observer()


        observer.schedule(BatchingWatchDogHandler(self.reload), MATERIALS_DIRECTORY, recursive=True)
        observer.start()

    def all_materials(self):
        """
        Return all loaded materials.
        """
        return list(self.materials.values())

    def delete_material(self, name):
        """
        Delete a specific material.
        """
        if name not in self.materials:
            raise KeyError(f"Material {name} not found")
        del self.materials[name]

    async def reload(self):
        log.info("Reloading materials ...")

        self.materials = {}

        paths = [path for paths_yielding_function in [
            list_files_in_resource_path(MATERIALS_CORE_RESOURCE),
            list_files_in_file_system(MATERIALS_DIRECTORY)
        ] for path in paths_yielding_function]

        for path in paths:
            filename = os.path.basename(path)
            if filename.endswith(".py"):
                # Import the file and execute material function to get the material
                module_name = os.path.splitext(filename)[0]
                spec = importlib.util.spec_from_file_location(
                    module_name, path)
                if not spec or spec.loader is None:
                    await connection_manager.send_message_to_all(NotificationWSMessage(title="Material not loaded", message=f"Skipping invalid material in file {filename}"))
                    continue

                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                material = module.material

                if not material or not isinstance(material, dict) or "usage" not in material:
                    await connection_manager.send_message_to_all(NotificationWSMessage(title="Material not loaded", message=f"Skipping invalid material in file {filename}"))
                    continue

                id = filename[:-3]
                if id in self.materials:
                    await connection_manager.send_message_to_all(NotificationWSMessage(title="Material not loaded", message=f"Skipping duplicate material {id} in file {filename}"))
                    continue

                if "content" not in material:
                    material["content"] = documentation_from_code(module_name, path)
            
                self.materials[id] = Material(
                    id=id, usage=material["usage"], content=material["content"]
                )
            elif filename.endswith(".md"):
                with open(path, "r") as file:
                    lines = file.readlines()

                    # Merging all lines into a single string
                    text = "".join(lines)

                    pattern = r"\s*(<!---|<!--)\s*(.*?)\s*(-->)\s*(.*)\s*"

                    match = re.match(pattern, text.strip(), re.DOTALL)

                    if not match:
                        await connection_manager.send_message_to_all(NotificationWSMessage(title="Material not loaded", message=f"Skipping invalid material in file {filename}"))
                        continue

                    # Extracting 'usage' and 'content' based on matched groups
                    usage = match.group(2)
                    content = match.group(4)

                    # Pruning leading/trailing spaces and newlines (if any)
                    usage = usage.strip()
                    content = content.strip()

                    material_id = os.path.splitext(filename)[0]
                    if material_id in self.materials:
                        await connection_manager.send_message_to_all(NotificationWSMessage(title="Material not loaded", message=f"Skipping duplicate material {material_id} in file {filename}"))
                        continue

                    self.materials[material_id] = Material(
                        id=material_id,
                        usage=usage,
                        content=lambda context, content=content: content,
                    )

        await connection_manager.send_message_to_all(NotificationWSMessage(title="Materials reloaded", message=f"Reloaded {len(self.materials)} materials"))
        


materials = Materials()
