import json
import logging
from fastapi import WebSocket
from typing import Dict, List, Literal
from collections import defaultdict

from pydantic import BaseModel

log = logging.getLogger(__name__)


class BaseWSMessage(BaseModel):
    type: str

class NotificationWSMessage(BaseWSMessage):
    type: Literal["notification"] = "notification"
    title: str
    message: str

class ErrorWSMessage(BaseWSMessage):
    type: Literal["error"] = "error"
    error: str

class ConnectionManager:
    def __init__(self):
        self.active_connections: Dict[str, List[WebSocket]] = defaultdict(list)

    async def connect(self, websocket: WebSocket, chat_id: str):
        await websocket.accept()
        log.info(f"Connected to chat {chat_id}")
        
        self.active_connections[chat_id].append(websocket)

    def disconnect(self, websocket: WebSocket, chat_id: str):
        self.active_connections[chat_id].remove(websocket)

    async def send_message(self, chat_id: str, msg: BaseWSMessage):
        log.info(f"Sending message to {chat_id}: {msg}")
        for connection in self.active_connections[chat_id]:
            await connection.send_text(msg.model_dump_json())

    async def send_message_to_all(self, msg: BaseWSMessage):
        log.info(f"Sending message to all: {msg}")
        for connections in self.active_connections.values():
            for connection in connections:
                await connection.send_text(msg.model_dump_json())

connection_manager = ConnectionManager()